////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// BeDVI                                                                                                          //
// by Achim Blumensath                                                                                            //
// blume@corona.oche.de                                                                                           //
//                                                                                                                //
// This program is free software! It may be distributed according to the GNU Public License (see COPYING).        //
//                                                                                                                //
// $Id: DVI-View.cc,v 1.4 1997/04/27 17:41:29 achim Exp achim $
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <InterfaceKit.h>
#include <stdio.h>
#include <string.h>
#include "debug.h"
#include "BeDVI.h"
#include "DVI-View.h"
#include "TeXFont.h"

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// DVIView::DVIView(BRect rect, const char *Name)                                                                 //
//                                                                                                                //
// Initialises a DVI-View.                                                                                        //
//                                                                                                                //
// BRect      rect                      bounds of the view                                                        //
// const char *Name                     name of the view                                                          //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

DVIView::DVIView(BRect rect, const char *Name):
  BView(rect, Name, B_FOLLOW_ALL_SIDES, B_WILL_DRAW | B_FRAME_EVENTS),
  DocLock(B_ERROR),
  Document(NULL),
  BufferLock(B_ERROR),
  Buffer(NULL),
  BufferView(NULL),
  PageNo(1)
{
  D(bug("DVIView::DVIView()\n"));

  if((DocLock = create_sem(1, NULL)) < B_NO_ERROR)
    return;

  if((BufferLock = create_sem(1, NULL)) < B_NO_ERROR)
    return;

  if(!(BufferView = new BView(BRect(0, 0, 1, 1), NULL, 0, 0)))
    return;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// DVIView::~DVIView()                                                                                            //
//                                                                                                                //
// Deletes a DVI-View.                                                                                            //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

DVIView::~DVIView()
{
  D(bug("DVIView::~DVIView()\n"));

  if(DocLock >= B_NO_ERROR)
  {
    acquire_sem(DocLock);

    if(Document)
      delete Document;
    
    delete_sem(DocLock);
  }
  else
    if(Document)
      delete Document;

  if(BufferLock >= B_NO_ERROR)
  {
    acquire_sem(BufferLock);

    if(Buffer)
      delete Buffer;

    else if(BufferView)           // deleted automatically if BBitmap is deleted
      delete BufferView;

    delete_sem(BufferLock);
  }
  else
  {
    if(Buffer)
      delete Buffer;

    else if(BufferView)           // deleted automatically if BBitmap is deleted
      delete BufferView;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::Draw(BRect r)                                                                                    //
//                                                                                                                //
// Draws the view (see the BeBook).                                                                               //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::Draw(BRect r)
{
  if(acquire_sem(DocLock) < B_NO_ERROR)
    return;

  if(Buffer != NULL &&
     acquire_sem_etc(BufferLock, 1, B_TIMEOUT, 0.0) == B_NO_ERROR)
  {
    DrawBitmapAsync(Buffer, r, r);

    release_sem(BufferLock);
  }
  else
    FillRect(Bounds(), B_SOLID_LOW);

  release_sem(DocLock);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::FrameResized(float Width, float Height)                                                          //
//                                                                                                                //
// Updates the scrollbars of the view (see the BeBook).                                                           //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::FrameResized(float Width, float Height)
{
  UpdateScrollBars();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::MessageReceived(BMessage *msg)                                                                   //
//                                                                                                                //
// Handles messages from the menu (see the BeBook).                                                               //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::MessageReceived(BMessage *msg)
{
  if(msg->WasDropped() && msg->HasRef("refs"))
  {
    entry_ref File;

    if(msg->FindRef("refs", &File) >= B_NO_ERROR)
      ((ViewApplication *)be_app)->OpenFile(File);
  }
  else
    switch(msg->what)
    {
      case MsgAntiAliasing:
        ToggleAntiAliasing();
        break;

      case MsgBorderLine:
        ToggleBorderLine();
        break;

      case MsgNext:
        IncPage();
        break;

      case MsgPrev:
        DecPage();
        break;

      case MsgFirst:
        SetPage(1);      
        break;

      case MsgLast:
        SetPage(0xffffffff);
        break;

      default:
        BView::MessageReceived(msg);
        break;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::KeyDown(const char *bytes, int32 numBytes)                                                       //
//                                                                                                                //
// Handles key presses (see the BeBook).                                                                          //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::KeyDown(const char *bytes, int32 numBytes)
{
  BRect  r;
  u_long w,h;
  float  MaxX, MaxY;
  float  Factor;

  if(!Document)
    return;

  r    = Bounds();
  w    = Document->PageWidth;
  h    = Document->PageHeight;
  MaxX = w - r.right  + r.left - 1.0;
  MaxY = h - r.bottom + r.top  - 1.0;

  Factor = 1.0 / 20.0;

  if(modifiers() & B_SHIFT_KEY)
    Factor *= 4.0;
  if(modifiers() & B_OPTION_KEY)
    Factor *= 0.25;

  switch(bytes[0])
  {
    case B_PAGE_DOWN:
    case B_SPACE:
      if(modifiers() & B_SHIFT_KEY)
        IncPage();
      else
      {
        if(r.top < MaxY)
          ScrollTo(r.left, min(r.bottom - 20.0, MaxY));
        else
        {
          if(IncPage())
            ScrollTo(r.left, 0.0);
        }
      }
      break;

    case B_PAGE_UP:
    case B_BACKSPACE:
      if(modifiers() & B_SHIFT_KEY)
        DecPage();
      else
      {
        if(r.top > 0.0)
          ScrollTo(r.left, max(2 * r.top - r.bottom + 20.0, 0.0));
        else
        {
          if(DecPage())
            ScrollTo(r.left, MaxY);
        }
      }
      break;

    case B_HOME:
      if(modifiers() & B_SHIFT_KEY)
        SetPage(1);
      else
        ScrollTo(r.left, 0.0);
      break;

    case B_END:
      if(modifiers() & B_SHIFT_KEY)
        SetPage(~0L);
      else
        ScrollTo(r.left, MaxY);
      break;

    case B_DOWN_ARROW:
    case B_RETURN:
      if(r.top < MaxY)
        ScrollTo(r.left, min(r.top + Factor * h, MaxY));
      else
      {
        if(IncPage())
          ScrollTo(r.left, 0.0);
      }
      break;

    case B_UP_ARROW:
      if(r.top > 0.0)
        ScrollTo(r.left, max(r.top - Factor * h, 0.0));
      else
      {
        if(DecPage())
          ScrollTo(r.left, MaxY);
      }
      break;

    case B_LEFT_ARROW:
      ScrollTo(max(r.left - Factor * w, 0.0), r.top);
      break;

    case B_RIGHT_ARROW:
      ScrollTo(min(r.left + Factor * w, MaxX), r.top);
      break;

    case '+':
      IncPage();
      break;

    case '-':
      DecPage();
      break;

    case '<':
      SetPage(1);
      break;

    case '>':
      SetPage(~0L);
      break;

    case 'a':
    case 'A':
      ToggleAntiAliasing();
      break;

    case 'b':
    case 'B':
      ToggleBorderLine();
      break;

    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
      ((ViewApplication *)be_app)->SetResolution(0, bytes[0] - '0');
      break;

    default:
      BView::KeyDown(bytes, numBytes);
      break;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::SetDocument(DVI *doc)                                                                            //
//                                                                                                                //
// Sets the document to be displayed.                                                                             //
//                                                                                                                //
// DVI *doc                             document                                                                  //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::SetDocument(DVI *doc)
{
  DVI   *OldDoc;
  u_int OldPageNo;

  D(bug("DVIView::SetDocument()\n"));
  assert(doc != NULL);

  if(acquire_sem(DocLock) < B_NO_ERROR)
  {
    delete doc;
    return;
  }

  try
  {
    OldDoc    = Document;
    OldPageNo = PageNo;

    Document = doc;
    PageNo   = 1;

    if(DocumentChanged())
      delete OldDoc;

    else
    {
      delete Document;

      Document = OldDoc;
      PageNo   = OldPageNo;
    }
  }
  catch(...)
  {
    D(bug("exception occurred at %s:%d!\n", __FILE__, __LINE__));
  }

  release_sem(DocLock);

  UpdateWindow();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool DVIView::SetPage(u_int no)                                                                                //
//                                                                                                                //
// Sets the displayed page of the document.                                                                       //
//                                                                                                                //
// u_int no                             page number                                                               //
//                                                                                                                //
// Result:                              `true' if the page number changed, `false' if it remains the same         //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool DVIView::SetPage(u_int no)
{
  char str[10];

  D(bug("DVIView::SetPage(%u)\n", no));

  if(acquire_sem(DocLock) < B_NO_ERROR)
    return false;

  if(no != PageNo && Document)
  {
    if(no < 1)
      no = 1;
    if(no > Document->NumPages)
      no = Document->NumPages;

    if(PageNo != no)
    {
      PageNo = no;

      RedrawBuffer();
    }
    else
      no = 0;
  }
  else
    no = 0;

  release_sem(DocLock);

  if(Window()->Lock())
  {
    // update page counter

    sprintf(str, "%u", PageNo);
    ((BTextControl *)Window()->FindView("PageNumber"))->SetText(str);

    Invalidate();
    Window()->Unlock();
  }

  if(no > 0)
    return true;
  else
    return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool DVIView::IncPage()                                                                                        //
//                                                                                                                //
// Displays the next page.                                                                                        //
//                                                                                                                //
// Result:                              `true' if the page number changed, `false' if it remains the same         //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool DVIView::IncPage()
{
  return SetPage(PageNo + 1);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool DVIView::DecPage()                                                                                        //
//                                                                                                                //
// Displays the previous page.                                                                                    //
//                                                                                                                //
// Result:                              `true' if the page number changed, `false' if it remains the same         //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool DVIView::DecPage()
{
  if(PageNo > 1)
    return SetPage(PageNo - 1);
  else
    return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::ToggleBorderLine()                                                                               //
//                                                                                                                //
// Draws or clears the lines indicating the border of the page.                                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::ToggleBorderLine()
{
  DVISettings *prefs;

  D(bug("DVIView::ToggleBorderLine()\n"));

  if(acquire_sem(DocLock) < B_NO_ERROR)
    return;

  if(Document)
    prefs = &Document->Settings;

  else
    prefs = &((ViewApplication *)be_app)->Settings;

  prefs->BorderLine = !prefs->BorderLine;

  ((BMenu *)Window()->FindView("Menu"))->FindItem(MsgBorderLine)->SetMarked(prefs->BorderLine);

  release_sem(DocLock);

  RedrawBuffer();

  if(Window()->Lock())
  {
    Invalidate();
    Window()->Unlock();
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::ToggleAntiAliasing()                                                                             //
//                                                                                                                //
// Draws or clears the lines indicating the border of the page.                                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::ToggleAntiAliasing()
{
  DVISettings *prefs;

  D(bug("DVIView::ToggleAntiAliasing()\n"));

  if(acquire_sem(DocLock) < B_NO_ERROR)
    return;

  if(Document)
    prefs = &Document->Settings;

  else
    prefs = &((ViewApplication *)be_app)->Settings;

  prefs->AntiAliasing = !prefs->AntiAliasing;

  ((BMenu *)Window()->FindView("Menu"))->FindItem(MsgAntiAliasing)->SetMarked(prefs->AntiAliasing);

  if(Document)
    Document->Fonts.FlushShrinkedGlyphes();

  release_sem(DocLock);

  RedrawBuffer();

  if(Window()->Lock())
  {
    Invalidate();
    Window()->Unlock();
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::UpdateScrollBars()                                                                               //
//                                                                                                                //
// Updates position and proportion of the scrollbars.                                                             //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::UpdateScrollBars()
{
  if(!Document)
  {
    ScrollBar(B_VERTICAL)->SetRange(0, 0);
    ScrollBar(B_HORIZONTAL)->SetRange(0, 0);
  }
  else
  {
    BRect  r;
    u_long Width;
    u_long Height;

    r = Bounds();

    Width  = r.right - r.left + 1;
    Height = r.bottom - r.top + 1;

    ScrollBar(B_VERTICAL)->  SetRange(0, Document->PageHeight - Height);
    ScrollBar(B_HORIZONTAL)->SetRange(0, Document->PageWidth  - Width);
    ScrollBar(B_VERTICAL)->  SetProportion((float)Height / Document->PageHeight);
    ScrollBar(B_HORIZONTAL)->SetProportion((float)Width  / Document->PageWidth);
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::UpdateWindowSize()                                                                               //
//                                                                                                                //
// Updates size and size limits of the window.                                                                    //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::UpdateWindowSize()
{
  BRect r1, r2;

  if(Document)
  {
    r1 = Window()->Bounds();
    r2 = Bounds();

    Window()->SetSizeLimits(120.0, Document->PageWidth + r1.right - r2.right + r2.left - r1.left - 1.0,
                            120.0, Document->PageHeight + r1.bottom - r2.bottom + r2.top - r1.top - 1.0);

    Window()->GetSizeLimits(&r2.left, &r2.right, &r2.top, &r2.bottom);

    if(r2.right < r1.right || r2.bottom < r1.bottom)
      Window()->ResizeTo(min(r1.right, r2.right) - r1.left + 1.0, min(r1.bottom, r2.bottom) - r1.top + 1.0);
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::UpdateWindow()                                                                                   //
//                                                                                                                //
// Updates the window completely.                                                                                 //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::UpdateWindow()
{
  D(bug("DVIView::UpdateWindow()\n"));

  if(Window()->Lock())
  {
    UpdateMenus();
    UpdateWindowSize();
    UpdateScrollBars();
    Invalidate();

    Window()->Unlock();
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::UpdateMenus()                                                                            //
//                                                                                                                //
// Updates the marked menu items.                                                                                 //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::UpdateMenus()
{
  DVISettings *prefs;
  BMenu       *Menu;
  int         i;

  if(Document)
    prefs = &Document->Settings;
  else
    prefs = &((ViewApplication *)be_app)->Settings;

  Menu = ((BMenu *)Window()->FindView("Menu"));

  for(i = 0; i < NumModes; i++)
    if(strcmp(prefs->DspInfo.Mode, ModeNames[i]) == 0)
    {
      Menu->FindItem(ModeDPI[i])->SetMarked(true);
      break;
    }

  Menu->FindItem(prefs->DspInfo.ShrinkFactor)->SetMarked(true);
  Menu->FindItem(MsgAntiAliasing)->SetMarked(prefs->AntiAliasing);
  Menu->FindItem(MsgBorderLine)->SetMarked(prefs->BorderLine);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::RedrawBuffer()                                                                                   //
//                                                                                                                //
// Redraws the page in the internal buffer.                                                                       //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::RedrawBuffer()
{
  if(Document)
  {
    SetSleepCursor();

    if(BufferView->Window()->Lock())
    {
      if(acquire_sem(BufferLock) == B_NO_ERROR)
      {
        Document->Draw(BufferView, PageNo);

        release_sem(BufferLock);
      }

      BufferView->Window()->Unlock();
    }
    SetNormalCursor();
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVIView::SetDspInfo(DisplayInfo *dsp)                                                                     //
//                                                                                                                //
// Sets display data.                                                                                             //
//                                                                                                                //
// DisplayInfo *dsp                     display information                                                       //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVIView::SetDspInfo(DisplayInfo *dsp)
{
  DVI *doc;

  assert(dsp != NULL);

  if(acquire_sem(DocLock) < B_NO_ERROR)
    return;

  if(Document)
  {
    SetSleepCursor();

    Document->SetDspInfo(dsp);

    DocumentChanged();

    SetNormalCursor();
  }

  release_sem(DocLock);

  UpdateWindow();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool DVIView::DocumentChanged()                                                                                //
//                                                                                                                //
// Updates the view when the document has changed. This procedure should be called with `DocLock' locked.         //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool DVIView::DocumentChanged()
{
  BRect r;

  D(bug("DVIView::DocumentChanged()\n"));

  if(Buffer)
    r = Buffer->Bounds();

  if(Buffer == NULL                                ||
     Document->PageWidth  != r.right  - r.left + 1 ||     // do we need to resize the buffer?
     Document->PageHeight != r.bottom - r.top  + 1)
  {
    BBitmap *NewBuffer = NULL;

    try
    {
      if(!(NewBuffer = new BBitmap(BRect(0, 0, Document->PageWidth - 1, Document->PageHeight - 1), B_COLOR_8_BIT, true)))
        return false;

      if(Buffer)
      {
        Buffer->RemoveChild(BufferView);                                  // don't delete view
        delete Buffer;
      }
      Buffer = NewBuffer;

      BufferView->ResizeTo(Document->PageWidth, Document->PageHeight);

      Buffer->AddChild(BufferView);
    }
    catch(...)
    {
      D(bug("exception occurred at %s:%d!\n", __FILE__, __LINE__));
      return false;
    }
  }

  RedrawBuffer();

  return true;
}
