////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// BeDVI                                                                                                          //
// by Achim Blumensath                                                                                            //
// blume@corona.oche.de                                                                                           //
//                                                                                                                //
// This program is free software! It may be distributed according to the GNU Public License (see COPYING).        //
//                                                                                                                //
// $Id: DVI.cc,v 1.4 1997/04/27 17:41:29 achim Exp achim $
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <InterfaceKit.h>
#include <StorageKit.h>
#include <stdio.h>
#include <syslog.h>

#include "debug.h"
#include "defines.h"

extern "C"
{
  #define string _string
  #include "kpathsea/magstep.h"
  #undef string
};

#include "BeDVI.h"
#include "DVI-View.h"
#include "DVI.h"
#include "FontList.h"
#include "TeXFont.h"

static const int NoMagStep = -29999;
static const int NoBuild   =  29999;


/* DrawInfo *******************************************************************************************************/


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// u_long DrawInfo::ReadInt(ssize_t Size)                                                                         //
//                                                                                                                //
// reads an integer of `Size' bytes.                                                                              //
//                                                                                                                //
// ssize_t Size                         size of the integer in bytes                                              //
//                                                                                                                //
// Result:                              read integer                                                              //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

u_long DrawInfo::ReadInt(ssize_t Size)
{
  u_long x = 0;

  assert(Size > 0 && Size <= sizeof(u_long));

  if(BufferPos == NULL)
    File->Read((u_char *)&x + sizeof(u_long) - Size, Size);
  else
  {
    memcpy((u_char *)&x + sizeof(u_long) - Size, BufferPos, Size);
    BufferPos += Size;
  }

  return x;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// long DrawInfo::ReadSInt(ssize_t Size)                                                                          //
//                                                                                                                //
// reads a signed integer of `Size' bytes.                                                                        //
//                                                                                                                //
// ssize_t Size                         size of the integer in bytes                                              //
//                                                                                                                //
// Result:                              read integer                                                              //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

long DrawInfo::ReadSInt(ssize_t Size)
{
  ssize_t ReadSize;
  long    x = 0;

  assert(Size > 0 && Size <= sizeof(long));

  if(BufferPos == NULL)
  {
    ReadSize = 1;
    File->Read((u_char *)&x + sizeof(long) - 1, ReadSize);
  }
  else
  {
    memcpy((u_char *)&x + sizeof(long) - 1, BufferPos, 1);
    BufferPos += 1;
  }

  if(x & 0x80)
    x |= 0xffffff00;

  x <<= (Size - 1) * 8;

  if(Size > 1)
    if(BufferPos == NULL)
    {
      ReadSize = Size - 1;
      File->Read((u_char *)&x + sizeof(long) - Size + 1, ReadSize);
    }
    else
    {
      memcpy((u_char *)&x + sizeof(long) - Size + 1, BufferPos, Size - 1);
      BufferPos += Size - 1;
    }

  return x;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DrawInfo::ReadString(char *Buffer, size_t len)                                                            //
//                                                                                                                //
// reads a string of `len' bytes to `Buffer'.                                                                     //
//                                                                                                                //
// char   *Buffer                       buffer the string is stored in                                            //
// size_t len                           length of the string in bytes                                             //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DrawInfo::ReadString(char *Buffer, size_t len)
{
  if(BufferPos == NULL)
  {
    File->Read(Buffer, len);
    Buffer[len] = 0;
  }
  else
  {
    memcpy(Buffer, BufferPos, len);
    Buffer[len] = 0;
    BufferPos += len;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DrawInfo::Skip(off_t off)                                                                                 //
//                                                                                                                //
// skips the specified number of bytes.                                                                           //
//                                                                                                                //
// off_t off                            number of bytes to skip                                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DrawInfo::Skip(off_t off)
{
  if(BufferPos == NULL)
    File->Seek(off, SEEK_CUR);
  else
    BufferPos += off;
}


/* DisplayInfo ****************************************************************************************************/


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DisplayInfo::SetResolution(u_int ppi, u_int sf)                                                           //
//                                                                                                                //
// Sets the resolution information.                                                                               //
//                                                                                                                //
// u_int ppi                            dpi value (0: don't change old value)                                     //
// u_int sf                             factor to shrink document (0: don't change old value)                     //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DisplayInfo::SetResolution(u_int ppi, u_int sf)
{
  int i;

  D(bug("DisplayInfo::SetResolution(%u,%u)\n", ppi, sf));

  if(ppi != 0)
  {
    // search for mode name

    for(i = NumModes - 1; i > 0; i--)
      if(ModeDPI[i] <= ppi)
        break;

    Mode          = ModeNames[i];
    PixelsPerInch = ppi;
  }
  if(sf != 0)
    ShrinkFactor  = sf;
};


/* DVI ************************************************************************************************************/


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// DVI::DVI(BPositionIO *File, DVISettings *prefs)                                                                //
//                                                                                                                //
// Initializes a DVI-Document.                                                                                    //
//                                                                                                                //
// BPositionIO *File                    file of the document                                                      //
// DVISettings *prefs                   various options                                                           //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

DVI::DVI(BPositionIO *File, DVISettings *prefs):
  Fonts(),
  DVIFile(File),
  Name(NULL),
  PageOffset(NULL),
  Magnification(1000),
  DimConvert(1.0),
  OffsetX(prefs->DspInfo.PixelsPerInch),
  OffsetY(prefs->DspInfo.PixelsPerInch),
  DisplayError(NULL)
{
  D(bug("DVI::DVI()\n"));

  assert(prefs != NULL);

  Settings = *prefs;

  if(!Reload())
  {
    if(Name)
      delete [] Name;
    if(PageOffset)
      delete [] PageOffset;

    delete DVIFile;
    DVIFile    = NULL;
    Name       = NULL;
    PageOffset = NULL;

    return;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// DVI::~DVI()                                                                                                    //
//                                                                                                                //
// Deletes a DVI-Document.                                                                                        //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

DVI::~DVI()
{
  D(bug("DVI::~DVI()\n"));

  if(Name)
    delete [] Name;
  if(PageOffset)
    delete [] PageOffset;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// int DVI::MagStepValue(float &mag)                                                                              //
//                                                                                                                //
// Adjusts a magnification value.                                                                                 //
//                                                                                                                //
// float &mag                           magnification                                                             //
//                                                                                                                //
// Result:                              adjusted value                                                            //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int DVI::MagStepValue(float &mag)
{
  int   Result;
  u_int dpi;

  acquire_sem(kpse_sem);
  dpi = kpse_magstep_fix((u_int)mag, Settings.DspInfo.PixelsPerInch, &Result);
  release_sem(kpse_sem);

  mag = (float)dpi;

  return Result ? Result : NoMagStep;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool DVI::Reload()                                                                                             //
//                                                                                                                //
// Reads general data of a DVI file.                                                                              //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool DVI::Reload()
{
  static const int BufferLen = 512;

  long    Numerator;
  long    Denominator;
  int     len;
  long    LastPageOffset;
  long    pos;
  u_char  *p;
  u_char  Buffer[BufferLen];
  ssize_t ReadSize;
  u_char  x;
  u_char  Command;
  int     i;

  try
  {
    D(bug("DVI::Reload()\n"));

    assert(DVIFile != NULL);

    DVIFile->Seek(0, SEEK_SET);

    if(ReadInt(DVIFile, 1) != Preamble)
      return false;
    if(ReadInt(DVIFile, 1) != 2)
      return false;

    Numerator     = ReadInt(DVIFile, 4);
    Denominator   = ReadInt(DVIFile, 4);
    Magnification = ReadInt(DVIFile, 4);
    DimConvert    = (((double)Numerator * Magnification) / ((double)Denominator * 1000.0));
    DimConvert    = DimConvert * (((long)Settings.DspInfo.PixelsPerInch) << 16) / 254000;
    TPicConvert   = Settings.DspInfo.PixelsPerInch * Magnification / 1000000.0;

    len = ReadInt(DVIFile, 1);

    if(!(Name = new char[len + 1]))
      return false;

    DVIFile->Read(Name, len);
    Name[len] = 0;

    D(bug("Numerator = %ld\nDenominator = %ld\nMagnification = %ld\nDimConvert = %g\nTPicConvert = %g\n",
          Numerator, Denominator, Magnification, DimConvert, TPicConvert));

    pos = DVIFile->Seek(0, SEEK_END);

    if(pos > BufferLen)
      pos -= BufferLen;
    else
      pos = 0;

    DVIFile->Seek(pos, SEEK_SET);

    ReadSize = BufferLen;
    DVIFile->Read(&Buffer, ReadSize);
    p = &Buffer[ReadSize];

    do
    {
      while(p > Buffer && *--p != Trailer)
        ;

      if(p <= Buffer)
        return false;
    }
    while(p[-1] != Trailer || p[-2] != Trailer || p[-3] != Trailer);

    pos += p - Buffer;

    for(x = *p; x == Trailer; x = ReadInt(DVIFile, 1))
      DVIFile->Seek(--pos, SEEK_SET);

    if(x != 2)
      return false;

    DVIFile->Seek(pos - 4, SEEK_SET);
    DVIFile->Seek((long)ReadInt(DVIFile, 4), SEEK_SET);

    if(ReadInt(DVIFile, 1) != Postamble)
      return false;

    LastPageOffset = ReadInt(DVIFile, 4);

    if(ReadInt(DVIFile, 4) != Numerator || ReadInt(DVIFile, 4) != Denominator || ReadInt(DVIFile, 4) != Magnification)
      return false;

    UnshrunkPageHeight = ((long)((long)ReadInt(DVIFile, 4) * DimConvert) >> 16) + 2 * Settings.DspInfo.PixelsPerInch;
    UnshrunkPageWidth  = ((long)((long)ReadInt(DVIFile, 4) * DimConvert) >> 16) + 2 * Settings.DspInfo.PixelsPerInch;

    D(bug("unshrunk page height = %ld\nunshrunk page width = %ld\n", UnshrunkPageHeight, UnshrunkPageWidth));

    ReadInt(DVIFile, 2);

    NumPages = ReadInt(DVIFile, 2);

    Fonts.FreeFonts();
    Fonts.FlushShrinkedGlyphes();

    for(Command = ReadInt(DVIFile, 1); Command >= FontDef1 && Command <= FontDef4; Command = ReadInt(DVIFile, 1))
      if(!Fonts.LoadFont(this, DVIFile, Command))
        if(DisplayError)
          (*DisplayError)("Font not found!");

    Fonts.FreeUnusedFonts();

    if(Command != PostPost)
      return false;

    if(PageOffset)
      delete [] PageOffset;

    if(!(PageOffset = new u_long[NumPages]))
      return false;

    PageOffset[NumPages - 1] = LastPageOffset;
    DVIFile->Seek(LastPageOffset, SEEK_SET);

    for(i = NumPages - 2; i >= 0; i--)
    {
      DVIFile->Seek(41, SEEK_CUR);
      DVIFile->Seek(PageOffset[i] = ReadInt(DVIFile, 4), SEEK_SET);
    }

    PageWidth  = (UnshrunkPageWidth  + Settings.DspInfo.ShrinkFactor - 1) / Settings.DspInfo.ShrinkFactor + 2;
    PageHeight = (UnshrunkPageHeight + Settings.DspInfo.ShrinkFactor - 1) / Settings.DspInfo.ShrinkFactor + 2;

    return true;
  }
  catch(...)
  {
    D(bug("exception occurred at %s:%d!\n", __FILE__, __LINE__));
    return false;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVI::SetDspInfo(DisplayInfo *dsp)                                                                         //
//                                                                                                                //
// Sets display data.                                                                                             //
//                                                                                                                //
// DisplayInfo *dsp                     display information                                                       //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVI::SetDspInfo(DisplayInfo *dsp)
{
  assert(dsp != NULL);

  Settings.DspInfo = *dsp;

  OffsetX = Settings.DspInfo.PixelsPerInch;
  OffsetY = Settings.DspInfo.PixelsPerInch;

  Reload();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVI::Special(DrawInfo *di, long len)                                                                      //
//                                                                                                                //
// Processes `special' commands (currently not implemented).                                                      //
//                                                                                                                //
// DrawInfo *di                         drawing information                                                       //
// long len                             length of the command                                                     //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVI::Special(DrawInfo *di, long len)
{
  static char *Cmd   = NULL;
  static long CmdLen = -1;
  char        *p;
  ssize_t     ReadSize;

  try
  {
    if(CmdLen < len)
    {
      if(Cmd)
        delete [] Cmd;

      if(Cmd = new char[len + 1])
        CmdLen = len;
      else
        CmdLen = -1;
    }
    if(Cmd)
    {
      ReadSize = len;
      di->ReadString(Cmd, len);
      Cmd[len] = 0;
    }
    else
      di->Skip(len);

    D(bug("special: `%s'\n", Cmd));

  // do something!?
  }
  catch(...)
  {
    D(bug("exception occurred at %s:%d!\n", __FILE__, __LINE__));
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVI::ChangeFont(u_long n)                                                                                 //
//                                                                                                                //
// Changes the current font.                                                                                      //
//                                                                                                                //
// u_long n                             number of the new font                                                    //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVI::ChangeFont(DrawInfo *di, u_long n)
{
  if(n < Fonts.TableLength())
    if((di->CurFont = Fonts[n]) != NULL)
    {
      di->MaxChar = di->CurFont->MaxChar;
      di->SetChar = di->CurFont->SetChar;
    }
    else
    {
      di->MaxChar = 255;
      di->SetChar = SetNoChar;
    }
  else
    if(DisplayError)
      (*DisplayError)("Non-existent font number!");
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVI::Draw(BView *vw, BRect r, u_int PageNo)                                                               //
//                                                                                                                //
// Draws a DVI-Document.                                                                                          //
//                                                                                                                //
// BView *vw                            view the document should be displayed in                                  //
// BRect r                              rectangle to be drawn                                                     //
// u_int PageNo                         page to be displayed                                                      //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVI::Draw(BView *vw, u_int PageNo)
{
  DrawInfo di;
  size_t   BufferLen;
  ssize_t  ReadSize;

  try
  {
    assert(vw != NULL);

    vw->SetDrawingMode(B_OP_COPY);

    vw->FillRect(vw->Bounds(), B_SOLID_LOW);

    if(Settings.AntiAliasing && Settings.DspInfo.ShrinkFactor > 1)
      vw->SetDrawingMode(B_OP_MIN);
    else
      vw->SetDrawingMode(B_OP_OVER);

    // read page into memory

    if(PageNo < NumPages)                                       // this is a little bit more than the actual page
      BufferLen = PageOffset[PageNo] - PageOffset[PageNo - 1];
    else
    {
      BufferLen =  DVIFile->Seek(0, SEEK_END);
      BufferLen -= PageOffset[PageNo - 1];
    }

    DVIFile->Seek(PageOffset[PageNo - 1], SEEK_SET);

    if(!(di.BufferPos = new u_char[BufferLen]))
      return;

    ReadSize = BufferLen;
    DVIFile->Read(di.BufferPos, ReadSize);

    di.Document   = this;
    di.vw         = vw;
    di.DspInfo    = &Settings.DspInfo;
    di.DimConvert = DimConvert;
    di.Virtual    = NULL;
    di.CurFont    = NULL;
    di.SetChar    = SetNoChar;
    di.File       = DVIFile;
    di.BufferEnd  = di.BufferPos + BufferLen;

    bzero(&di.Data, sizeof(di.Data));

    DrawPart(&di);

    delete [] (di.BufferEnd - BufferLen);

    if(Settings.BorderLine)
    {
      BRect   r;
      pattern Dash = {0xf0, 0xf0, 0xf0, 0xf0, 0x0f, 0x0f, 0x0f, 0x0f};

      vw->SetDrawingMode(B_OP_COPY);

      r.Set(Settings.DspInfo.PixelsPerInch / Settings.DspInfo.ShrinkFactor,
            Settings.DspInfo.PixelsPerInch / Settings.DspInfo.ShrinkFactor,
            PageWidth  - Settings.DspInfo.PixelsPerInch / Settings.DspInfo.ShrinkFactor,
            PageHeight - Settings.DspInfo.PixelsPerInch / Settings.DspInfo.ShrinkFactor);

      vw->StrokeLine(BPoint(r.left,  0.0),      BPoint(r.left,        PageHeight - 1), Dash);
      vw->StrokeLine(BPoint(r.right, 0.0),      BPoint(r.right,       PageHeight - 1), Dash);
      vw->StrokeLine(BPoint(0.0,     r.top),    BPoint(PageWidth - 1, r.top),          Dash);
      vw->StrokeLine(BPoint(0.0,     r.bottom), BPoint(PageWidth - 1, r.bottom),       Dash);
    }

    vw->Sync();
  }
  catch(...)
  {
    D(bug("exception occurred at %s:%d!\n", __FILE__, __LINE__));
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVI::DrawPart(DrawInfo *di)                                                                               //
//                                                                                                                //
// Draws a part of a document.                                                                                    //
//                                                                                                                //
// DrawInfo *di                         drawing information                                                       //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVI::DrawPart(DrawInfo *di)
{
  u_char c;

  while(true)
  {
    c = di->ReadInt(1);

    if(c <= (u_char)(SetChar0 + 127))
      di->Data.Horiz += (*di->SetChar)(di, c);

    else if(FontNum0 <= c && c <= (u_char)(FontNum0 + 63))
      ChangeFont(di, (u_long)(c - FontNum0));

    else
    {
      long a, b;

      switch(c)
      {
        case Set1:
          di->Data.Horiz += (*di->SetChar)(di, di->ReadInt(1));
          break;

        case Put1:
          (*di->SetChar)(di, di->ReadInt(1));
          break;

        case SetRule:
          a = di->ReadInt(4) * DimConvert;
          b = di->ReadInt(4) * DimConvert;

          if(a > 0 && b > 0)
            DrawRule(di, Settings.DspInfo.PixelConv(di->Data.Horiz), di->Data.PixelV - Settings.DspInfo.ToPixel(a) + 1,
                         Settings.DspInfo.ToPixel(b),                Settings.DspInfo.ToPixel(a));

          di->Data.Horiz += b;
          break;

        case PutRule:
          a = di->ReadInt(4) * DimConvert;
          b = di->ReadInt(4) * DimConvert;

          if(a > 0 && b > 0)
            DrawRule(di, Settings.DspInfo.PixelConv(di->Data.Horiz), di->Data.PixelV - Settings.DspInfo.ToPixel(a) + 1,
                         Settings.DspInfo.ToPixel(b), Settings.DspInfo.ToPixel(a));
          break;

        case NOP:
          break;

        case BeginOP:
          di->Skip(44);

          di->Data.Horiz  = OffsetX << 16;
          di->Data.Vert   = OffsetY << 16;
          di->Data.PixelV = Settings.DspInfo.PixelConv(di->Data.Vert);
          di->Data.w      = 0;
          di->Data.x      = 0;
          di->Data.y      = 0;
          di->Data.z      = 0;

          break;

        case EndOP:
          if(!di->FrameList.Empty())
            if(DisplayError)
              (*DisplayError)("Stack not empty at EOP!");
          return;

        case Push:
        {
          Frame *f;

          if(!di->FreeFrames.Empty())
            f = (Frame *)di->FreeFrames.RemHead();
          else
            f = new Frame();

          if(!f)
          {
            if(DisplayError)
              (*DisplayError)("Not enough memory!");
            return;
          }
          f->Data = di->Data;
          di->FrameList.AddTail(f);
          break;
        }
        case Pop:
        {
          Frame *f;

          if(di->FrameList.Empty())
          {
            if(DisplayError)
              (*DisplayError)("Stack empty at Pop!");
            return;
          }

          f = (Frame *)di->FrameList.RemTail();

          di->Data = f->Data;
          di->FreeFrames.AddTail(f);
          break;
        }
        case Right1:
        case Right2:
        case Right3:
        case Right4:
          di->Data.Horiz += di->ReadSInt(c - Right1 + 1) * DimConvert;
          break;

        case W1:
        case W2:
        case W3:
        case W4:
          di->Data.w      = di->ReadSInt(c - W0) * DimConvert;
        case W0:
          di->Data.Horiz += di->Data.w;
          break;

        case X1:
        case X2:
        case X3:
        case X4:
          di->Data.x      = di->ReadSInt(c - X0) * DimConvert;
        case X0:
          di->Data.Horiz += di->Data.x;
          break;

        case Down1:
        case Down2:
        case Down3:
        case Down4:
          di->Data.Vert  += di->ReadSInt(c - Down1 + 1) * DimConvert;
          di->Data.PixelV = Settings.DspInfo.PixelConv(di->Data.Vert);
          break;

        case Y1:
        case Y2:
        case Y3:
        case Y4:
          di->Data.y      = di->ReadSInt(c - Y0) * DimConvert;
        case Y0:
          di->Data.Vert  += di->Data.y;
          di->Data.PixelV = Settings.DspInfo.PixelConv(di->Data.Vert);
          break;

        case Z1:
        case Z2:
        case Z3:
        case Z4:
          di->Data.z      = di->ReadSInt(c - Z0) * DimConvert;
        case Z0:
          di->Data.Vert  += di->Data.z;
          di->Data.PixelV = Settings.DspInfo.PixelConv(di->Data.Vert);
          break;

        case Font1:
        case Font2:
        case Font3:
        case Font4:
          ChangeFont(di, di->ReadInt(c - Font1 + 1));
          break;

        case XXX1:
        case XXX2:
        case XXX3:
        case XXX4:
          a = di->ReadInt(c - XXX1 + 1);

          if(a > 0)
            Special(di, a);
          break;

        case FontDef1:
        case FontDef2:
        case FontDef3:
        case FontDef4:
          di->Skip(13 + c - FontDef1);
          di->Skip(di->ReadInt(1) + di->ReadInt(1));
          break;

        default:
          D(bug("invalid operand: 0x%02x!\n", (u_int)c));

          if(DisplayError)
            (*DisplayError)("Invalid operand!");
          return;
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// long DVI::SetNoChar(DrawInfo *di, u_char c)                                                                    //
//                                                                                                                //
// Draws a character if currently no font is selected.                                                            //
//                                                                                                                //
// DrawInfo *di                         drawing information                                                       //
// u_char   c                           character to be drawn                                                     //
//                                                                                                                //
// Result:                              amount the cursor moves to the right or 0 if an error occures             //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

long DVI::SetNoChar(DrawInfo *di, u_char c)
{
  if(di->Virtual)
  {
    di->CurFont = di->Virtual->FirstFont;

    if(di->CurFont != NULL)
    {
       di->SetChar = di->CurFont->SetChar;

       return (*di->SetChar)(di, c);
    }
  }

  return 0;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// long DVI::SetVFChar(DrawInfo *di, u_char c)                                                                    //
//                                                                                                                //
// Draws a character if currently a virtual font is selected.                                                     //
//                                                                                                                //
// DrawInfo *di                         drawing information                                                       //
// u_char   c                           character to be drawn                                                     //
//                                                                                                                //
// Result:                              amount the cursor moves to the right or 0 if an error occures             //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

long DVI::SetVFChar(DrawInfo *di, u_char c)
{
  Macro         *m;
  DrawInfo      NewInfo;
  static u_char ch;

  if((m = &di->CurFont->Macros[c])->Position == NULL)
  {
    m->Position = &ch;
    m->End      = &ch;
    return 0;
  }

  NewInfo.Document     = di->Document;
  NewInfo.vw           = di->vw;
  NewInfo.Data         = di->Data;
  NewInfo.Data.w       = 0;
  NewInfo.Data.x       = 0;
  NewInfo.Data.y       = 0;
  NewInfo.Data.z       = 0;
  NewInfo.DspInfo      = di->DspInfo;
  NewInfo.DimConvert   = di->DimConvert;
  NewInfo.CurFont      = di->CurFont;
  NewInfo.VirtTable    = &di->CurFont->VFTable;
  NewInfo.Virtual      = di->Virtual;
  NewInfo.MaxChar      = di->MaxChar;
  NewInfo.SetChar      = di->SetChar;
  NewInfo.BufferPos    = m->Position;
  NewInfo.BufferEnd    = m->End;

  while(!di->FrameList.Empty())
    NewInfo.FrameList.AddTail(di->FrameList.RemHead());

  while(!di->FreeFrames.Empty())
    NewInfo.FreeFrames.AddTail(di->FreeFrames.RemHead());

  di->Document->DrawPart(&NewInfo);

  while(!NewInfo.FrameList.Empty())
    di->FrameList.AddTail(NewInfo.FrameList.RemHead());

  while(!NewInfo.FreeFrames.Empty())
    di->FreeFrames.AddTail(NewInfo.FreeFrames.RemHead());

  return m->Advance;  
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// long DVI::SetChar(DrawInfo *di, u_char c)                                                                      //
//                                                                                                                //
// Draws a character if currently a normal font is selected.                                                      //
//                                                                                                                //
// DrawInfo *di                         drawing information                                                       //
// u_char   c                           character to be drawn                                                     //
//                                                                                                                //
// Result:                              amount the cursor moves to the right or 0 if an error occures             //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

long DVI::SetChar(DrawInfo *di, u_char c)
{
  Glyph *g;

  try
  {
    if(c > di->MaxChar)
      return 0;

    g = &di->CurFont->Glyphs[c];

    if(g->UBitMap == NULL)
      di->CurFont->ReadChar(di->CurFont, c);

    if(g->UBitMap == NULL)
      return 0;

    if(di->DspInfo->ShrinkFactor == 1)
      di->vw->DrawBitmapAsync(g->UBitMap, BPoint(di->DspInfo->PixelConv(di->Data.Horiz) - g->Ux, di->Data.PixelV - g->Uy));

    else
    {
      g->Shrink(di->DspInfo->ShrinkFactor, di->Document->Settings.AntiAliasing);

      di->vw->DrawBitmapAsync(g->SBitMap, BPoint(di->DspInfo->PixelConv(di->Data.Horiz) - g->Sx, di->Data.PixelV - g->Sy));
    }
    return g->Advance;
  }
  catch(...)
  {
    D(bug("exception occurred at %s:%d!\n", __FILE__, __LINE__));
    return 0;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVI::DrawRule(DrawInfo *di, long x, long y, long w, long h)                                               //
//                                                                                                                //
// Draws a rule.                                                                                                  //
//                                                                                                                //
// DrawInfo *di                         drawing information                                                       //
// long     x, y, w, h                  size of the rule                                                          //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVI::DrawRule(DrawInfo *di, long x, long y, long w, long h)
{
  di->vw->FillRect(BRect(x, y, x + (w ? w : 1) - 1, y + (h ? h : 1) - 1), B_SOLID_HIGH);
}
