////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// BeDVI                                                                                                          //
// by Achim Blumensath                                                                                            //
// blume@corona.oche.de                                                                                           //
//                                                                                                                //
// This program is free software! It may be distributed according to the GNU Public License (see COPYING).        //
//                                                                                                                //
// $Id: FontList.cc,v 1.4 1997/04/27 17:41:29 achim Exp achim $
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <StorageKit.h>
#include <string.h>
#include <syslog.h>
#include "debug.h"
#include "BeDVI.h"
#include "FontList.h"
#include "TeXFont.h"


/* FontList *******************************************************************************************************/


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// FontList::FontList()                                                                                           //
//                                                                                                                //
// Initializes a FontList.                                                                                        //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

FontList::FontList()
{
  ListLock = create_sem(1, NULL);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// FontList::~FontList()                                                                                          //
//                                                                                                                //
// Deletes a FontList.                                                                                            //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

FontList::~FontList()
{
  if(ListLock >= B_NO_ERROR)
  {
    acquire_sem(ListLock);
    delete_sem(ListLock);
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// Font *FontList::LoadFont(DVI *doc, const char *Name, float Size, long ChkSum, int MagStep, double DimConvert)  //
//                                                                                                                //
// Loads a font and adds it to the list.                                                                          //
//                                                                                                                //
// DVI        *doc                      document the font appears in                                              //
// const char *Name                     name of the font                                                          //
// float      Size                      size of the font                                                          //
// long       ChkSum                    checksum                                                                  //
// int        MagStep                   magnification                                                             //
// double     DimConvert                factor to convert dimensions                                              //
//                                                                                                                //
// Result:                              pointer to the font or `NULL' if an error occures                         //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Font *FontList::LoadFont(DVI *doc, const char *Name, float Size, long ChkSum, int MagStep, double DimConvert)
{
  Font *f;

  D(bug("FontList::LoadFont(%s, %f, %ld, %d, %g)\n", Name, Size, ChkSum, MagStep, DimConvert));

  if(acquire_sem(ListLock) < B_NO_ERROR)
    return NULL;

  try
  {
    for(f = (Font *)Fonts.First; f->Next; f = (Font *)f->Next)
      if(strcmp(f->Name, Name) == 0 && (int)(Size + 0.5) == (int)(f->Size + 0.5))
        break;

    if(f->Next == NULL)
    {
      if(!(f = new Font(doc, Name, Size, ChkSum, MagStep, DimConvert)))
      {
        release_sem(ListLock);
        return NULL;
      }
      if(!f->Loaded)
      {
        release_sem(ListLock);

        delete f;
        return NULL;
      }

      Fonts.AddTail(f);
    }

    atomic_add(&f->UseCount, 1);
  }
  catch(...)
  {
    D(bug("exception occurred at %s:%d!\n", __FILE__, __LINE__));
    if(f)
      delete f;
    f = NULL;
  }
  release_sem(ListLock);
  
  return f;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void FontList::FreeFont(Font *f)                                                                               //
//                                                                                                                //
// frees a font.                                                                                                  //
//                                                                                                                //
// Font *f                              font to be freed                                                          //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void FontList::FreeFont(Font *f)
{
  if(atomic_add(&f->UseCount, -1) == 1)
  {
    if(acquire_sem(ListLock) < B_NO_ERROR)
      return;

    f->Remove();

    release_sem(ListLock);

    delete f;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void FontList::FreeAll()                                                                                       //
//                                                                                                                //
// Frees all fonts in the list.                                                                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void FontList::FreeAll()
{
  if(acquire_sem(ListLock) < B_NO_ERROR)
    return;

  Fonts.Clear();

  release_sem(ListLock);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void FontList::FreeUnusedFonts()                                                                               //
//                                                                                                                //
// Frees all unused fonts in the list.                                                                            //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void FontList::FreeUnusedFonts()
{
  Font *f;

  if(acquire_sem(ListLock) < B_NO_ERROR)
    return;

  for(f = (Font *)Fonts.First; f->Next; f = (Font *)f->Next)
    if(f->UseCount < 1)
    {
      f = (Font *)f->Prev;

      delete f->Next->Remove();
    }

  release_sem(ListLock);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void FontList::FlushShrinkedGlyphes()                                                                          //
//                                                                                                                //
// Frees all bitmaps of shrinked glyphes which is neccessary if the anti aliasing method is changed.              //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void FontList::FlushShrinkedGlyphes()
{
  Font *f;

  if(acquire_sem(ListLock) < B_NO_ERROR)
    return;

  for(f = (Font *)Fonts.First; f->Next; f = (Font *)f->Next)
    f->FlushShrinkedGlyphes();

  release_sem(ListLock);
}


/* FontTable ******************************************************************************************************/


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// FontTable::FontTable()                                                                                         //
//                                                                                                                //
// Initializes a FontTable.                                                                                       //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

FontTable::FontTable():
  TableSem(B_ERROR),
  Table(NULL),
  TableLen(0)
{
  if((TableSem = create_sem(1, NULL)) < B_NO_ERROR)
    return;

  if(Table = new Font *[16])
  {
    TableLen = 16;
    bzero(Table, TableLen * sizeof(Font *));
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// FontTable::~FontTable()                                                                                        //
//                                                                                                                //
// Deletes a FontTable.                                                                                           //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

FontTable::~FontTable()
{
  if(TableSem >= B_NO_ERROR)
  {
    acquire_sem(TableSem);
    delete_sem(TableSem);
  }

  if(Table)
  {
    FreeFonts();
    delete [] Table;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// Font *FontTable::LoadFont(DVI *doc, BPositionIO *File, u_char Command)                                         //
//                                                                                                                //
// Loads a font and adds it to the table.                                                                         //
//                                                                                                                //
// DVI         *doc                     document the font appears in                                              //
// BPositionIO *File                    file which contains the fontname                                          //
// u_char      Command                  Font-Definition command                                                   //
//                                                                                                                //
// Result:                              pointer to the font or `NULL' if an error occured                         //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Font *FontTable::LoadFont(DVI *doc, BPositionIO *File, u_char Command)
{
  Font    *NewFont;
  int     TeXNo;
  long    ChkSum;
  int     Scale;
  int     Design;
  int     MagStep;
  int     len;
  char    *FontName;
  ssize_t ReadSize;
  float   FontSize;
  int     Size;
  double  ScaleDimConvert;

  try
  {
    TeXNo  = ReadInt(File, Command - FontDef1 + 1);
    ChkSum = ReadInt(File, 4);
    Scale  = ReadInt(File, 4);
    Design = ReadInt(File, 4);

    len    = ReadInt(File, 1);
    len   += ReadInt(File, 1);

    if(!(FontName = new char[len + 1]))
      return NULL;

    ReadSize = len;
    File->Read(FontName, ReadSize);
    FontName[len] = 0;

    FontSize        = 0.001 * Scale / Design * doc->Magnification * doc->Settings.DspInfo.PixelsPerInch;
    ScaleDimConvert = doc->DimConvert;

    MagStep = doc->MagStepValue(FontSize);
    Size    = FontSize + 0.5;

    if(!(NewFont = Fonts.LoadFont(doc, FontName, FontSize, ChkSum, MagStep, Scale * ScaleDimConvert / (double)(1L << 20))))
    {
      delete [] FontName;
      return NULL;
    }

    delete [] FontName;

    if(TeXNo >= TableLen)
      if(!Resize(TeXNo + 8))
        return NULL;

    Table[TeXNo] = NewFont;

    return NewFont;
  }
  catch(...)
  {
    D(bug("exception occurred at %s:%d!\n", __FILE__, __LINE__));
    return NULL;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool FontTable::Resize(u_long len)                                                                             //
//                                                                                                                //
// Resizes the table.                                                                                             //
//                                                                                                                //
// u_long len                           new length of the table                                                   //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool FontTable::Resize(u_long len)
{
  Font **NewTable;

  try
  {
    D(bug("FontTable::Resize(%lu)\n", len));

    if(!(NewTable = new Font *[len]))
      return false;

    memcpy(NewTable, Table, min(len, TableLen) * sizeof(Font *));

    if(len > TableLen)
      bzero(NewTable + TableLen, (len - TableLen) * sizeof(Font *));

    if(acquire_sem(TableSem) != B_NO_ERROR)
    {
      delete [] NewTable;
      return false;
    }

    if(Table)
      delete [] Table;

    Table    = NewTable;
    TableLen = len;

    release_sem(TableSem);

    return true;
  }
  catch(...)
  {
    D(bug("exception occurred at %s:%d!\n", __FILE__, __LINE__));
    return false;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void FontTable::FreeFonts()                                                                                    //
//                                                                                                                //
// Frees all fonts in the table.                                                                                  //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void FontTable::FreeFonts()
{
  int i;

  for(i = TableLen - 1; i >= 0; i--)
    if(Table[i])
    {
      Fonts.FreeFont(Table[i]);
      Table[i] = NULL;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void FontTable::FreeUnusedFonts()                                                                              //
//                                                                                                                //
// Frees all unused fonts in the list.                                                                            //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void FontTable::FreeUnusedFonts()
{
  Fonts.FreeUnusedFonts();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void FontTable::FlushShrinkedGlyphes()                                                                         //
//                                                                                                                //
// Frees all bitmaps of shrinked glyphes which is neccessary if the anti aliasing method is changed.              //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void FontTable::FlushShrinkedGlyphes()
{
  Fonts.FlushShrinkedGlyphes();
}
